/**
 * AuxIdx.java
 * Una clase que contiene las funciones auxiliares
 * utilizadas por las funciones geométricas
 * @version 1.0, 04/01/12
 * @author Santiago González
 */

import java.util.*;
import java.sql.*;
import org.h2.tools.SimpleResultSet;
import gnu.trove.TIntStack;


public class AuxIdx {
    
	public static int nRoot;
        public static RTree arbolR[] = new RTree[1000];
        public static String arbol[]= new String[1000]; 
        private static int numarbol=0;

     /**
      * Función que realiza la creación de la tabla persistente con el índice espacial
      * para , posteriormente, realizar su carga en memoria dentro de la lista de árboles
      * Rtree del sistema.
      * @param conn Conexión con la base de datos
      * @param tableDb nombre de la tabla que contiene el campo geométrico a indexar
      * @param columnTb nombre del campo geométrico a indexar
      */
   public static boolean createSpatialIndex(Connection conexion, String tableDb, String columnTb) throws Exception {

		// Comprobamos incialmente si el campo geométrico ya existe en las tablas de metadatos
		Statement stmt = conexion.createStatement();
		ResultSet rs = stmt.executeQuery("SELECT count(*) as cont from JASPA.GEOMETRY_COLUMNS WHERE (F_TABLE_NAME='"+ tableDb.toUpperCase()+ "' or F_TABLE_NAME='" + tableDb.toLowerCase() +"') and (F_GEOMETRY_COLUMN='" + columnTb.toUpperCase()+ "' or F_GEOMETRY_COLUMN='" + columnTb.toLowerCase()+"')");
		rs.next();
		int num = rs.getInt("cont");
		stmt.close();
		
		// En caso de que el campo ya exista (num<0)comenzamos el proceso de indexación
		if (num>0) {

			// Creamos la tabla de índice persistente a partir del nombre de tabla y de campo geométrico.
			try {							
				Statement stmt3 = conexion.createStatement();	
				stmt3.execute("DROP TABLE IF EXISTS SpatialDataMBRTable_" + tableDb +"_" +columnTb);
				stmt3.close();
                                Statement stmt13 = conexion.createStatement();	
                                stmt13.execute("CREATE TABLE SpatialDataMBRTable_" + tableDb +"_" +columnTb + "(SGID INT PRIMARY KEY,XMIN REAL, YMIN REAL, XMAX REAL, YMAX REAL)");
				stmt13.close();
                                System.out.println();
				System.out.println("La estructura de la tabla indice se ha creado correctamente y se llama: SpatialDataMBRTable_" + tableDb + "_" + columnTb);
				System.out.println();
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}						
			
			// Insertamos en esta nueva tabla los MBR de los valores pertenecientes al campo geométrico a indexar
			
                        insertMBRFromSpatialDataTableIntoSpatialDataMBRTable(conexion, tableDb, columnTb);
			System.out.println();
			System.out.println("La tabla indice se ha rellenado convenientemente con los MBR de los datos geometricos y se han cargado en memoria");
			System.out.println();				
							
			// Se crean los disparadores que mantendrán el índice actualizado tanto en memoria como en tablas persistentes
			try {							
				Statement stmt4 = conexion.createStatement();						
				stmt4.execute("DROP ALIAS IF EXISTS TRIGGER_SET_" + tableDb + "_" +columnTb + ";CREATE ALIAS TRIGGER_SET_" + tableDb +"_" +columnTb + " FOR \"" + TriggerIdx.class.getName() + ".setTriggerData\"");						
				stmt4.close();	
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}

			//Se crea un trigger para actualizar el índice al insertar un registro en la tabla de datos
			
			try {							
				Statement stmt5 = conexion.createStatement();						
				stmt5.execute("DROP TRIGGER IF EXISTS " + tableDb +"_" +columnTb + "_INS; CREATE TRIGGER "+ tableDb +"_" +columnTb +"_INS AFTER INSERT ON " + tableDb + " FOR EACH ROW CALL \"TriggerIdx\" ");	
				stmt5.close();	
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}

			//Se crea un trigger para actualizar el índice al eliminar un registro en la tabla de datos
			try {								
				Statement stmt6 = conexion.createStatement();						
				stmt6.execute("DROP TRIGGER IF EXISTS " + tableDb +"_" +columnTb + "_DEL; CREATE TRIGGER "+ tableDb +"_" +columnTb +"_DEL AFTER DELETE ON " + tableDb + " FOR EACH ROW CALL \"TriggerIdx\" ");		
				stmt6.close();			
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}							
				
			//Se crea un trigger para actualizar el índice al actualizar un registro en la tabla de datos

			try {	
				Statement stmt7 = conexion.createStatement();						
				stmt7.execute("DROP TRIGGER IF EXISTS " + tableDb +"_" +columnTb + "_UPD; CREATE TRIGGER "+ tableDb +"_" +columnTb +"_UPD AFTER UPDATE ON " + tableDb + " FOR EACH ROW CALL \"TriggerIdx\" ");	
				stmt7.close();						
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}	
                        
			try {	
				Statement stmt8 = conexion.createStatement();						
				stmt8.execute("CALL TRIGGER_SET_"+ tableDb +"_" +columnTb +"('"+ tableDb +"_" +columnTb + "_INS','"+columnTb+"')");				
				stmt8.close();						
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}							
				
			try {							
				Statement stmt9 = conexion.createStatement();						
				stmt9.execute("CALL TRIGGER_SET_"+ tableDb +"_" +columnTb +"('"+ tableDb +"_" +columnTb + "_DEL','"+columnTb+"')");
				stmt9.close();						
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}							
				
			try {								
				Statement stmt10 = conexion.createStatement();						
				stmt10.execute("CALL TRIGGER_SET_"+ tableDb +"_" +columnTb +"('"+ tableDb +"_" +columnTb + "_UPD','"+columnTb+"')");	
				stmt10.close();						
			} catch(SQLException ex) {
				System.err.println("SQLException: " + ex.getMessage());
			}	
                        

			System.out.println();
			System.out.println("Los triggers asociados a la tabla de datos y a la tabla indice se han creado y activado correctamente");
			System.out.println();		

				
			return true;
		}

		// Se devuelve un mensaje de error en caso de que el campo no exista dentro de las tablas de metadatos
		else {
			System.out.println("Error: La tabla a indexar no existe o la columna a indexar no existe o no es de tipo geometrico");	
		    return false;			
		}
	}
   
   
    /**
      * Función que realiza la inserción en la tabla persistente y en el árbol RTree en memoria del índice espacial
      * @param conn Conexión con la base de datos
      * @param tableDb nombre de la tabla que contiene el campo geométrico a indexar
      * @param columnTb nombre del campo geométrico a indexar
      */

       public static void insertMBRFromSpatialDataTableIntoSpatialDataMBRTable(Connection conexion, String tableDb, String columnTb)
	throws Exception {		
                   
	// Comprueba la posicion del campo geometrico dentro de la estructura de la tabla		
	ResultSet rs2 = conexion.getMetaData().getColumns(null,null,tableDb,null);
	int pos = 0;
        String column="";
        String iDColName="";
		
	// Se recorren todos los campos de la tabla para comprobar si contiene un campo de igual nombre que el pasado como argumento
	while (rs2.next()) {
		if (pos==0) iDColName = rs2.getString("COLUMN_NAME");	
			column = rs2.getString("COLUMN_NAME");
			if (column.equals(columnTb)) break;
			else pos++;
		}

		// El primer campo de la tabla que contiene los datos geometricos debe ser la clave
		Statement stmt = conexion.createStatement();
		ResultSet conjuntoResultados = stmt.executeQuery("select " + iDColName + " ,ST_xmin(" + columnTb + "),ST_ymin(" + columnTb + "),ST_xmax(" + columnTb + "),ST_ymax(" + columnTb + ")  from " + tableDb);
                

		// Se crea una nueva estructura de árbol Rtree
		arbolR[numarbol] = new RTree();
		Properties 	props = new Properties();
		props.setProperty("maxNodeEntries","6");
		props.setProperty("minNodeEntries","3");
		
		// Se actualizan las propiedades de la estructura árbol creada.
		arbolR[numarbol].init(props);
	
		int nSigNodos=20;
		int i;
		int k;
		int h;
		String l;
		int nLevel=0;
		int maxLevel=0;
		int childNumber=0;
		int childId=0;
		Node newNode = null;
		Node raizNode = null;
		Node childNode = null;
			
		i=1 ;	
		Statement ps = conexion.createStatement();		
		while(conjuntoResultados.next()) {
				 
			 float x0,y0,x1,y1;
			 
			 // Se obtiene el identificador de cada registro
			 int id = Integer.valueOf(conjuntoResultados.getObject(1).toString()).intValue() ;
						 
			 // Se calcula el MBR de un campo geométrico
			 x0 = Float.valueOf(conjuntoResultados.getObject(2).toString()).floatValue() ;
			 y0 = Float.valueOf(conjuntoResultados.getObject(3).toString()).floatValue() ;
			 x1 = Float.valueOf(conjuntoResultados.getObject(4).toString()).floatValue() ;
			 y1 = Float.valueOf(conjuntoResultados.getObject(5).toString()).floatValue() ;
						  
			// Se genera un nuevo valor para el árbolR a partir de los datos espaciales
			arbolR[numarbol].add(new Rectangle(x0,y0,x1,y1),id);
			
			// También se va cargando en la tabla persistente del indice los MBR de los datos que se van leyendo de la tabla de datos
                        
                        ps.execute("INSERT INTO SpatialDataMBRTable_"+ tableDb + "_" + columnTb + " VALUES(" +Integer.toString(id) + ","+ Float.toString(x0) + ","+ Float.toString(y0)+ ","+Float.toString(x1)+","+Float.toString(y1)+")");										 
		}

                System.out.println("Insertados datos para "+tableDb+"','"+columnTb+".");
                
                // Se actualiza la tabla persistente con la lista de índices espaciales en memoria
                ps.execute("INSERT INTO ACT_IDX VALUES ('"+tableDb+"','"+columnTb+"',"+numarbol+")");
                arbol[numarbol]=tableDb.concat(columnTb);
                numarbol++;
                System.out.println("Árbol para '"+tableDb+" , "+columnTb+"' creado e insertado");
                        
        stmt.close();		
		ps.close();		
	
	}


     /**
      * Función que obtiene los objetos totalmente contenidos dentro de un rectángulo dado
      * @param conn Conexión con la base de datos
      * @param tableDb nombre de la tabla que contiene el campo geométrico a indexar
      * @param columnTb nombre del campo geométrico a indexar
      * @param XMin,YMin,XMax,YMax valores del rectángulo MBR sobre el que se realiza la búsqueda de elementos
      * @return devuelve el identificador de los distintos elementos incluidos dentro del MBR de entrada
      */

	public static SimpleResultSet getContainFunction(Connection conn, String tableName, String field,float Xmin, float Ymin, float Xmax,float Ymax) throws Exception {	

		

               // Se obtiene la posición dentro de la lista de índices espaciales sobre la que se realiza la búsqueda
                int pos;
                for (pos=0;(!arbol[pos].equals(tableName.concat(field)));pos++);
                
                // Se crea la estructura de los registros que almacenará los identificadores de los registros
		// como resultado de la consulta
                SimpleResultSet rs4 = new SimpleResultSet();
                rs4.addColumn("ID", Types.INTEGER, 10, 0);
		
                
		/**
                 * Implementación de la búsqueda de todos los objetos contenidos completamente
		 * dentro de un rectángulo dado.
		 * Se introducen en el array IDsIn[] todos los SRID de los objetos que están contenidos dentro
		 * del rectánguo que se pide.
		 * El numero de SRID devueltos es IDsCont. Si no hay ninguno, queda IDsCont=0
                 */

		int startIndex=1;
		int IDsIn[] = new int[1000000];
		int IDsCont=0;
		int i,j,k,h;
		TIntStack parents = new TIntStack();
		TIntStack parentsEntry = new TIntStack();
		                
		nRoot=0;
		int nlevel=0;

            for (int m=0; m <= arbolR[pos].getHighestUsedNodeId();m++){
                        Node newNode = arbolR[pos].getNode(m);

			if (newNode.level>nlevel){
				nlevel=newNode.level;
				nRoot=m;
			}
		}
		parents.reset();
		parents.push(nRoot);	
		
		parentsEntry.reset();
		parentsEntry.push(-1);
		Node n = null;
		IDsCont=0;
		Node nRaiz = arbolR[pos].getNode(nRoot);
		while (parents.size() > 0){
                    n = arbolR[pos].getNode(parents.peek());
			//n = arbolR[rs.getInt(1)].getNode(parents.peek());
			startIndex = parentsEntry.peek() + 1;
			if (!n.isLeaf()){
				boolean contains = false;
				for (i = startIndex; i < n.entryCount; i++){
					if (Rectangle.intersects(Xmin, Ymin, Xmax, Ymax, n.entriesMinX[i],
						n.entriesMinY[i], n.entriesMaxX[i],n.entriesMaxY[i])){
						parents.push(n.ids[i]);
						parentsEntry.pop();
						parentsEntry.push(i);
						parentsEntry.push(-1);
						contains=true;
						break;
					};
				};
				if (contains) { continue; };
			} else{
					for (h = 0; h < n.entryCount; h++){
						if ( (Xmin <= n.entriesMinX[h]) && (Ymin <= n.entriesMinY[h]) &&
							(Xmax >= n.entriesMaxX[h]) && (Ymax >= n.entriesMaxY[h]) ){
							// Encontrado objeto
							IDsCont++;
							IDsIn[IDsCont]=n.ids[h];
					};
			};
			};
			parents.pop();
			parentsEntry.pop();
		};

		// Presenta los objetos encontrados o una indicacion de que no se ha localizado
		if (IDsCont>0){
			for (k=1; k<=IDsCont; k++){
 				rs4.addRow(IDsIn[k]);
			};
		} 

		// Devuelve los identificadores (en forma de registros) que cumplen los criterios de la consulta espacial realizada	
		return rs4;

	}
        
     /**
      * Función que obtiene los objetos total o parcialmente contenidos dentro de un rectángulo dado
      * @param conn Conexión con la base de datos
      * @param tableDb nombre de la tabla que contiene el campo geométrico a indexar
      * @param columnTb nombre del campo geométrico a indexar
      * @param XMin,YMin,XMax,YMax valores del rectángulo MBR sobre el que se realiza la búsqueda de elementos
      * @return devuelve el identificador de los distintos elementos incluidos dentro del MBR de entrada
      */

	public static SimpleResultSet getIntersectFunction(Connection conexion, String tableName, String field,float Xmin, float Ymin, float Xmax,float Ymax) throws Exception {	

		// Se obtiene la posición dentro de la lista de índices espaciales sobre la que se realiza la búsqueda
                int pos=0;
                for (pos=0;(!arbol[pos].equals(tableName.concat(field)));pos++);

		// Se crea la estructura de los registros que almacenará los identificadores de los registros
		// como resultado de la consulta
                
                SimpleResultSet rs4 = new SimpleResultSet();
                rs4.addColumn("ID", Types.INTEGER, 10, 0);
		
		/**
                 * Implementación de la búsqueda de todos los objetos contenidos completamente
		 * dentro de un rectángulo dado.
		 * Se introducen en el array IDsIn[] todos los SRID de los objetos que están contenidos dentro
		 * del rectánguo que se pide.
		 * El numero de SRID devueltos es IDsCont. Si no hay ninguno, queda IDsCont=0
                 */

		int startIndex=1;
		int IDsIn[] = new int[1000000];
		int IDsCont=0;
		int i,j,k,h;
		TIntStack parents = new TIntStack();
		TIntStack parentsEntry = new TIntStack();

		nRoot=0;
		int nlevel=0;
                for (int m=0; m <= arbolR[pos].getHighestUsedNodeId();m++){
			Node newNode = arbolR[pos].getNode(m);
 			if (newNode.level>nlevel){
				nlevel=newNode.level;
				nRoot=m;
			}
		}

		parents.reset();
		parents.push(nRoot);
	
		parentsEntry.reset();
		parentsEntry.push(-1);
		Node n = null;
		IDsCont=0;
		Node nRaiz = arbolR[pos].getNode(nRoot);
		
		while (parents.size() > 0){
                    n = arbolR[pos].getNode(parents.peek());
			startIndex = parentsEntry.peek() + 1;
			if (!n.isLeaf()){
				boolean intersects = false;
				for (i = startIndex; i < n.entryCount; i++){
					if (Rectangle.intersects(Xmin, Ymin, Xmax, Ymax, n.entriesMinX[i],
						n.entriesMinY[i], n.entriesMaxX[i],n.entriesMaxY[i])){
						parents.push(n.ids[i]);
						parentsEntry.pop();
						parentsEntry.push(i);
						parentsEntry.push(-1);
						intersects=true;
						break;
					};
				};
				if (intersects) { continue; };
			} else{
					for (h = 0; h < n.entryCount; h++){
					if (Rectangle.intersects(Xmin, Ymin, Xmax, Ymax, n.entriesMinX[h],
						n.entriesMinY[h], n.entriesMaxX[h],n.entriesMaxY[h])){
							// Encontrado objeto
							IDsCont++;
							IDsIn[IDsCont]=n.ids[h];
					};
			};
			};
			parents.pop();
			parentsEntry.pop();
		};

		// Presenta los objetos encontrados o una indicacion de que no se ha localizado
		if (IDsCont>0){
			for (k=1; k<=IDsCont; k++){
				rs4.addRow(IDsIn[k]);
			};
		} 

		// Devuelve los identificadores (en forma de registros) que cumplen los criterios de la consulta espacial realizada	
		return rs4;
    }	
        
                
     /**
      * Función que compara dos rectángulos MBR dados a través de sus respectivos índices espaciales
      * y devuelve si interseccionan entre ellos o no
      * Se utilizan las funciones de JSI modificadas para adaptarse a nuestra necesidad
      * @param conn Conexión con la base de datos
      * @param val valor del MBR a comparar
      * @param tableName nombre de la tabla que contiene el campo geométrico
      * @param field nombre del campo geométrico
      * @param val2 valor del segundo MBR a comparar
      * @param tableName2 nombre de la tabla que contiene el campo geométrico 2
      * @param field2 nombre del campo geométrico 2
      * @return devuelve true en caso de que los MBR tengan intersección entre ellos o false en caso contrario
      */
	public static boolean getIntersectMBRFunction(Connection conexion, String tableName, String field, Rectangle val,String tableName2, String field2,Rectangle val2) throws Exception {


                // Se obtiene la posición dentro de la lista de índices espaciales sobre la que se realiza la búsqueda para MBR1
                int pos=0;
                for (pos=0;(!arbol[pos].equals(tableName.concat(field)));pos++);

                // Se obtiene la posición dentro de la lista de índices espaciales sobre la que se realiza la búsqueda para MBR2
                int pos2=0;
                for (pos2=0;(!arbol[pos2].equals(tableName2.concat(field2)));pos2++);

                // Se obtienen los nodos de los índices respectivos que contienen los MBR
                Node n = null;
                Node n2 = null;
                n = arbolR[pos].getNode(findLeaf_idx(val, pos));
                n2 = arbolR[pos2].getNode(findLeaf_idx(val2, pos2));

                // Se obtiene la posición de los MBR dentro de cada nodo
                int a = findEntry_idx(val,n);
                int b = findEntry_idx(val2,n2);

                // Se comparan los valores de los MBR para comprobar si interseccionan entre ellos
                if (Rectangle.intersects(n.entriesMinX[a],n.entriesMinY[a], n.entriesMaxX[a],n.entriesMaxY[a], n2.entriesMinX[b],
						n2.entriesMinY[b], n2.entriesMaxX[b],n2.entriesMaxY[b])){
					return true;
					}
                 else {
                    return false;}
                 
    }	
        
     /**
      * Función que devuelve la posición de una geometría dentro de un nodo del índice espacial
      * @param r MBR que se está buscando en el sistema
      * @param n nodo en el que se realiza la búsqueda del elemento
      * @return devuelve la posición del MBR en el nodo o -1 en caso de que no exista
      */
    private static int findEntry_idx(Rectangle r, Node n) {
        for (int i = 0; i < n.entryCount; i++) {
            if (n.entriesMinX[i] == r.minX && n.entriesMinY[i] == r.minY &&
                n.entriesMaxX[i] == r.maxX && n.entriesMaxY[i] == r.maxY) {
            return i;
            }
        }
        return -1;
  }
   

     /**
      * Función que devuelve el nodo hoja dentro de un índice espacial en el que se encuentra un MBR dado
      * @param r MBR que se está buscando en el sistema
      * @param pos posición del índice espacial en el que se realiza la búsqueda
      * @return devuelve la posición del MBR en el nodo o -1 en caso de que no exista
      */
    public static int findLeaf_idx(Rectangle r, int pos) {

	TIntStack parents = new TIntStack();
	TIntStack parentsEntry = new TIntStack();
        Rectangle r2 = new Rectangle();

	nRoot=0;
	int nlevel=0;
        for (int m=0; m <= arbolR[pos].getHighestUsedNodeId();m++){
		Node newNode = arbolR[pos].getNode(m);
 		if (newNode.level>nlevel){
			nlevel=newNode.level;
			nRoot=m;
		}
	}
		
	parents.reset();
	parents.push(nRoot);
  	
  	parentsEntry.reset();
  	parentsEntry.push(-1);
  	Node n = null;
  	int foundIndex = -1;  
  	
  	while (foundIndex == -1 && parents.size() > 0) {
  	  n = arbolR[pos].getNode(parents.peek());
  	  int startIndex = parentsEntry.peek() + 1;
      
      if (!n.isLeaf()) {
	  	  boolean contains = false;
        for (int i = startIndex; i < n.entryCount; i++) {
	  	    if (Rectangle.contains(n.entriesMinX[i], n.entriesMinY[i], n.entriesMaxX[i], n.entriesMaxY[i],
                                 r.minX, r.minY, r.maxX, r.maxY)) { 
	  	      parents.push(n.ids[i]);
	  	      parentsEntry.pop();
	  	      parentsEntry.push(i); 
	  	      parentsEntry.push(-1);
	  	      contains = true;
            break; 
	  	    }
	  	  }
        if (contains) {
          continue;
        }
      }
      else{
          for (int h = 0; h < n.entryCount; h++){
             
                r2.set(n.entriesMinX[h],n.entriesMinY[h], n.entriesMaxX[h],n.entriesMaxY[h]);
		if (r.equals(r2)){
							// Encontrado objeto
							foundIndex=n.nodeId;
					};
          }
      }
      parents.pop();
      parentsEntry.pop();
  	} 
 		return foundIndex;
  	}
   
   
     /**
      * Función que ejecuta la adición o la eliminación de un elemento en un índice espacial
      * @param inst define si es una inserción (1) o una actualización (4)
      * @param postree posición del índice espacial al que afecta la operación
      * @param gid identificador del elemento que va a ser añadido o eliminado
      * @param r MBR del elemento geométrico que va a ser añadido o eliminado
      */
	public static void updateIDX(int inst, int postree, int gid, Rectangle r)  {	
            
            if (inst==1){
                arbolR[postree].add(r, gid);
            }

            
            else if (inst==4){
                arbolR[postree].delete(r, gid);
            }
        }

     /**
      * Función que realiza el lanzamiento de las acciones de creación automáticas de los índices espaciales en el sistema
      * @param conn Conexión con la base de datos
      */
	public  static void initIDX(Connection conn) throws Exception {	
            

                Statement stmt = conn.createStatement();
                // inicializamos las tablas persistentes de los índices espaciales
                stmt.execute("DROP TABLE IF EXISTS ACT_IDX;CREATE TABLE ACT_IDX (TABLENAME VARCHAR(30), FIELD VARCHAR(30), POSITION INT, PRIMARY KEY(TABLENAME,FIELD))");

                //Obtenemos los índices activos en el sistema
                ResultSet indices = stmt.executeQuery("SELECT F_TABLE_NAME as tableName,F_GEOMETRY_COLUMN field from JASPA.GEOMETRY_COLUMNS");
                
                int i=1;
                // actualizamos las tablas persistentes de los índices espaciales
                while (indices.next() && i<=5){
                  createSpatialIndex(conn,indices.getString(1),indices.getString(2));
                   i++;
    
                }
            }
}